# Native imports
import json
import os
from builtins import FileNotFoundError
from pathlib import Path

# 3rd party imports
import jsonpickle

def get_known_items(knownitems_path: Path) -> dict:
    """Checks if item ID is in JSON file's list of known items"""
    # TODO Detect when message has been sent to Slack queue but not added to
    # the list of known modqueue items
    try:
        with open(knownitems_path, 'r') as itemfile:
            jsonstr = f'{itemfile.read()}'
            knownitems = jsonpickle.decode(jsonstr)
    except (json.JSONDecodeError, FileNotFoundError):
        knownitems = {}
    return knownitems

def find_latest(message_ts: str, post_dir: Path) -> str:
    """Retrieves the latest POST request timestamp for a given message."""
    latest_ts = message_ts
    for postfile in os.listdir(os.fsencode(post_dir)):
        if (filename := os.fsdecode(postfile)).endswith('.json'):
            request_ts = filename.strip('.json')
            if request_ts < latest_ts: 
                continue
            else:
                with open(os.path.join(post_dir, filename), 'r') as file:
                    request = json.load(file)
                if request['container']['message_ts'] == message_ts:
                    if request_ts > latest_ts : latest_ts = request_ts
                else:
                    continue
        else:
            continue
    return latest_ts

def update_knownitems_file(knownitems: dict, knownitems_path: Path):
    with open(knownitems_path, 'w+') as itemfile:
        jsonstr = jsonpickle.encode(knownitems)
        print(jsonstr, file=itemfile)

def clean_post_request(incomplete_items: dict, postrequest_path: Path):
    """Remove POST request files for completed items"""
    keepjson_ts = []
    for item in incomplete_items.values():
        requests, timestamps = item.find_post_requests(postrequest_path)
        if requests:
            for request in requests:
                keepjson_ts.append(request)
    for postfile in postrequest_path.iterdir():
        if postfile.stem not in keepjson_ts:
            os.remove(postfile)

def cleanup_knownitems_json(incomplete_items: dict, knownitems_path: Path):
    """Clean known item JSON"""
    with open(knownitems_path, 'w+') as itemfile:
        jsonstr = jsonpickle.encode(incomplete_items)
        print(jsonstr, file=itemfile)



# Future imports
from __future__ import (
    annotations
)

# Standard imports
import argparse
from typing import (
    Sequence
)
from pathlib import Path

# Local imports
import reddack
import reddack.config
import reddack.utils

def create_arg_parser() -> argparse.ArgumentParser:
    """Create the argument parser for the CLI"""

    parser = argparse.ArgumentParser(
        description=(
            "Moderate Reddit communities via Slack"
        ),
        argument_default=argparse.SUPPRESS
    )

    parser.add_argument(
        "--config",
        dest="config_path",
        required=True,
        help="The path to the config file."
    )

    parser.add_argument(
        "--queue",
        action="store_true"
    )

    return parser

def process_args(parsedargs):
    configpath = Path(parsedargs.configpath)
    if configpath.suffix == ".json":
        reddack_objs = reddack.config.reddack_from_file(configpath)
    if parsedargs.queue:
        for objs in reddack_objs:
            reddack.utils.sync(objs)

def cli(sys_argv: Sequence[str] | None = None) -> None:
    """Parse the CLI arguments"""
    parser = create_arg_parser()
    parsed_args = parser.parse_args(sys_argv)
    process_args(parsed_args)

def main(sys_argv: Sequence[str] | None = None) -> None:
    """Run through the CLI."""
    cli(sys_argv)



from __future__ import annotations
import abc
from typing import (
    Type,
    Union,
    TypedDict
)
from pathlib import Path
import json

# 3rd party imports
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
from praw import Reddit
from praw.models import Submission
import jsonpickle

# Local imports
from reddack.payload import (
    build_removal_block,
    build_submission_block, 
    build_response_block, 
    build_archive_blocks
)
from reddack.exceptions import (
    MsgSendError
)
from reddack.slack import (
    ReddackResponse,
    SubmissionResponse
)
from reddack.utils import (
    get_known_items,
    update_knownitems_file,
    cleanup_knownitems_json,
    clean_post_request
)

# TODO Add functionality for flairing posts
# TODO Add functionality for awarding posts

class ReddackItem:
    """Stores information about the state of an item in the modqueue."""
    def __init__(self, prawitem):
        self.prawitem = prawitem.id
        self.message_ts = None
        self.responses: dict[str, ReddackResponse] = {}

    def process_slack_responses(self, postrequest_path: Path): 
        """Check for responses to mod item message."""
        requests, timestamps = self.find_post_requests(postrequest_path)
        if requests:
            for request, timestamp in zip(requests, timestamps):
                moderator = request['user']['id']
                if moderator in self.responses:
                    self.responses[moderator].update(request, timestamp)
                else:
                    self.initialize_response(moderator)
                    self.responses[moderator].update(request, timestamp)

    def find_post_requests(self, postrequest_path: Path) -> tuple[list[dict], list[str]]:
        """Retrieves all POST requests for a given mod item, returns sorted list."""
        recieved_timestamps = []
        requests = []
        for postfile in postrequest_path.iterdir():
            if postfile.suffix == '.json':
                request_ts = postfile.stem
                with open(postfile, 'r') as file:
                    request = json.load(file)
                if request['container']['message_ts'] == self.message_ts:
                    recieved_timestamps.append(request_ts)
                    requests.append(request)
                else:
                    continue
            else:
                continue
        if len(recieved_timestamps) == 0:
            return ((), ())
        else:
            sortedrequests, sortedtimestamps = zip(
                *sorted(zip(requests, recieved_timestamps), 
                key=lambda z: z[1])
            )
            return sortedrequests, sortedtimestamps

class ReddackComment(ReddackItem):
    """Stores information about the state of a comment in the modqueue."""

class ReddackSubmission(ReddackItem):
    """Stores information about the state of a submission in the modqueue."""

    _ResponseType : Type = SubmissionResponse
    kind : str = "submission"

    def __init__(self, prawitem):
        self.created_utc = prawitem.created_utc
        self.title = prawitem.title
        self.url = prawitem.url
        self.author = prawitem.author.name
        self.thumbnail = prawitem.thumbnail
        self.text = prawitem.selftext
        self.permalink = prawitem.permalink
        super().__init__(prawitem)

    def send_msg(self, client: WebClient, channel: str, removal_options: dict):
        """Send message for new mod item to specified Slack channel"""
        try:
            result = client.chat_postMessage(
                blocks=self.msg_payload(removal_options), channel=channel, 
                text="New modqueue item", unfurl_links=False, unfurl_media=False
            )
            result.validate()
            self.message_ts = result.data['ts']
        except SlackApiError as error:
            try:
                result = client.chat_postMessage(
                    blocks=self.NoneType(removal_options, thumbnail=False), channel=channel, 
                    text="New modqueue item", unfurl_links=False, unfurl_media=False
                )
                result.validate()
                self.message_ts = result.data['ts']
            except SlackApiError as error:
                raise MsgSendError("Failed to send item to Slack.") from error

    def _delete_msg(self, client: WebClient, user_client: WebClient, channel: str):
        """Delete replies to mod item message"""
        response = client.conversations_replies(
            channel=channel, 
            ts=self.message_ts
        )
        for message in response["messages"][::-1]:
            reply_response = user_client.chat_delete(
                channel=channel, 
                ts=message["ts"],
                as_user=True
            )

    def _send_archive(self, client: WebClient, channel: str):
        """Send archive message after mod actions are complete"""
        responseblocks = []
        for userid, modresponse in self.responses.items():
            response = client.users_info(user=userid)
            name = response["user"]["real_name"]
            responseblocks.append(
                build_response_block(
                    name, 
                    modresponse.states["actionApproveRemove"].value, 
                    modresponse.states["actionRemovalReason"].value
                )
            )
        archiveblocks = build_archive_blocks(
            self.created_utc, 
            self.title,
            self.author,
            self.permalink,
            responseblocks,
        )
        with open("debugdump.json", "w+") as f:
            blocksjson = jsonpickle.encode(archiveblocks)
            print(blocksjson, file=f)
        result = client.chat_postMessage(
            blocks=archiveblocks, channel=channel,
            text="Archived modqueue item", unfurl_links=False, unfurl_media=False
        )

    def complete_cleanup(self, client: WebClient, user_client: WebClient, channels: str):
        """Delete message and send to archive after completion"""
        self._send_archive(client, channels['archive'])
        self._delete_msg(client, user_client, channels['queue'])

    def initialize_response(self, moderator: str):
        """Initialize a new moderator response object"""
        self.responses[moderator] = self._ResponseType(self.message_ts)
    
    def approve_or_remove(self, thresholds: Thresholds) -> str | None:
        """Decide whether to approve or remove item based on moderator votes and stored thresholds"""
        votesum = 0
        for response in self.responses.values():
            if response.actions['actionConfirm'].value: 
                votesum += float(response.states['actionApproveRemove'].value)
        if votesum >= thresholds['approve']:
            return 'approve'
        elif votesum <= thresholds['remove']:
            return "remove"
        else:
            return None
    
    def msg_payload(self, removal_options: dict, thumbnail: bool=True) -> dict:
        "Create Slack message payload for the Submission"
        try:
            return build_submission_block(
                self.created_utc, 
                self.title, 
                self.url, 
                self.author, 
                self.thumbnail if thumbnail else thumbnail,
                self.text,
                self.permalink,
                removal_options

            )
        except AttributeError as error:
            raise MsgSendError(
                f"{error.obj!r} object is missing field {error.name!r}."
            )

    @property
    def removal_reasons(self):
        unique_reasons = []
        for response in self.responses.values():
            for reason in response.states['actionRemovalReason'].value:
                if reason not in unique_reasons: unique_reasons.append(reason)
        return sorted(unique_reasons)

    @property
    def modnote(self):
        modnote = ""
        timestamp = "0"
        for response in self.responses.values():
            if response.states['actionModnote'].timestamp > timestamp:
                modnote = response.states['actionModnote'].value
        return modnote
                

class Auth(abc.ABC):
    @abc.abstractmethod
    def create_client(self):
        pass

class PrawAuth(Auth):

    AGENT = "r/SpaceX Slack moderation interface by u/ModeHopper"

    def __init__(self, 
        client_id: str, 
        client_secret: str, 
        username: str, 
        password: str
    ):
        self.client_id = client_id
        self.client_secret = client_secret
        self.username = username
        self.password = password

    def create_client(self) -> Reddit:
        """Create an instance of praw.Reddit using stored authenticators"""
        return Reddit(
            client_id = self.client_id,
            client_secret = self.client_secret,
            password = self.password,
            username = self.username,
            user_agent = self.AGENT
        )

class SlackAuth(Auth):
    def __init__(self, bot_token: str, user_token: str):
        self.bot_token = bot_token
        self.user_token = user_token

    def create_client(self, as_user: bool =False) -> WebClient:
        """Create an instance of the slack_sdk.WebClient using stored authenticators"""
        return WebClient(token=(self.user_token if as_user else self.bot_token))

class Thresholds(TypedDict):
    approve: int
    remove: int

class Channels(TypedDict):
    queue: str
    archive: str

class RemovalTemplate(TypedDict):
    pre: str
    post: str

class Rule(TypedDict):
    title: str
    text: str
    shorttext: str
    link: str
    applyto: str

class Reddack:
    def __init__(self,
        subreddit_name: str,
        praw_auth: PrawAuth,
        slack_auth: SlackAuth,
        channels: dict[ReddackItem, Channels],
        rules: dict[str, Rule],
        knownitems_path: Path = Path.cwd() / 'KNOWN_ITEMS.json',
        postrequest_path: Path = Path.cwd() / 'POST',
        thresholds: dict[ReddackItem, Thresholds] = {
            ReddackSubmission: {
                'approve': +1,
                'remove': -1
            },
            ReddackComment: {
                'approve': +1,
                'remove': -1
            }
        },
        removal_template: RemovalTemplate = None
    ):
        self.subreddit_name = subreddit_name
        self.praw_auth = praw_auth
        self.slack_auth = slack_auth
        self.knownitems_path = knownitems_path
        self.postrequest_path = postrequest_path
        self.rules = rules
        self.channels = channels
        self.thresholds = thresholds
        self.removal_template = removal_template
        self.removal_options = build_removal_block(self.rules)

    @property
    def subreddit(self):
        return self.praw_auth.create_client().subreddit(self.subreddit_name)

    @property
    def praw_client(self):
        return self.praw_auth.create_client()

    @property
    def slack_client(self):
        return self.slack_auth.create_client()

    @property
    def slack_user_client(self):
        return self.slack_auth.create_client(as_user=True)

    @property
    def subreddit(self):
        return self.praw_client.subreddit(self.subreddit_name)
    
    def sync(self):
        """Sync the modqueue between Slack and Reddit"""
        knownitems = get_known_items(self.knownitems_path)
        newitems = self.check_reddit_queue(knownitems)
        knownitems = self.update_slack_queue(newitems, knownitems)
        update_knownitems_file(knownitems, self.knownitems_path)
        knownitems = self.check_slack_queue(knownitems)
        update_knownitems_file(knownitems, self.knownitems_path)

    def check_reddit_queue(self, 
        knownitems: dict[str, ReddackItem]
    ) -> dict[str, ReddackItem]:
        """Check Reddit modqueue for unmoderated items"""
        newitems = {}
        for item in self.subreddit.mod.modqueue(limit=None):
            # Check if item is comment or submission
            if isinstance(item, Submission):
                ReddackItem = ReddackSubmission
            else:
                print("Ignoring non-submission item.")
                continue
            # Check if item is known
            isknown = True if item.id in knownitems else False
            if isknown:
                continue
            else:
                newitems[item.id] = ReddackItem(item)
        return newitems

    def update_slack_queue(self, 
        newitems: dict[str, ReddackItem], 
        knownitems: dict[str, ReddackItem]
    ) -> dict[str, ReddackItem]:
        """Update the Slack modqueue with new Reddit modqueue items"""
        # TODO Detect when an item has been removed from the Reddit modqueue and remove it from Slack modqueue
        for id, moditem in newitems.items():
            moditem.send_msg(
                self.slack_client, 
                self.channels[type(moditem)]['queue'],
                self.removal_options
            )
            # Add to known items
            knownitems[id] = moditem
        return knownitems
        
    def send_removal_message(self, moditem: ReddackItem):
        """Send a modmail with the selected removal reasons and optional modnote"""
        body = self.removal_template["pre"]
        for removal_reason in moditem.removal_reasons:
            rule = self.rules[removal_reason]
            body +=f"\n\n>{rule['title']}: {rule['text']}"
        if moditem.modnote:
            body += f"\n\n**Moderator note:** {moditem.modnote}"
        body += "\n\n" + self.removal_template["post"]
        body += f"\n\n**Content of your {moditem.kind}**: {moditem.title}\n\n>"
        body += moditem.url if not moditem.text else moditem.text
        body += f"\n\n---\n\n[Link to your submission]({moditem.permalink})"
        subject = f"Your {moditem.kind} has been removed from r/{self.subreddit_name}"
        conversation = self.subreddit.modmail.create(subject, body, moditem.author)
        conversation.archive()
        
    def check_slack_queue(self, 
        knownitems: dict[str, ReddackItem]
    ) -> dict[str, ReddackItem]:
        """Check Slack items for new moderation actions"""
        incomplete = {}
        complete = {}
        if knownitems is None:
            knownitems = {}
        for moditem in knownitems.values():
            moditem.process_slack_responses(self.postrequest_path)
            if moditem.approve_or_remove(self.thresholds[type(moditem)]) == "approve":
                self.praw_client.submission(moditem.prawitem).mod.approve()
            elif moditem.approve_or_remove(self.thresholds[type(moditem)]) == "remove":
                self.praw_client.submission(moditem.prawitem).mod.remove()
                self.send_removal_message(moditem)
            else:
                incomplete[moditem.prawitem] = moditem
                continue
            complete[moditem.prawitem] = moditem
            moditem.complete_cleanup(
                self.slack_client, 
                self.slack_user_client, 
                self.channels[type(moditem)]
            )
        cleanup_knownitems_json(incomplete, self.knownitems_path)
        clean_post_request(incomplete, self.postrequest_path)
        knownitems = incomplete
        return knownitems

    



from datetime import datetime, timezone


# Dictionary of month names
months = {
    1: 'January', 2: 'February', 3: 'March',
    4: 'April', 5: 'May', 6: 'June', 7: 'July',
    8: 'August', 9: 'September', 10: 'October',
    11: 'November', 12: 'December'
}

# Lambda function for converting cardinal to ordinal
ordinal = lambda n : "%d%s" % (n,"tsnrhtdd"[(n//10%10!=1)*(n%10<4)*n%10::4])


def build_response_block(name: str, vote, removalreasons: list) -> dict:
    """Build blocks for moderator responses in archive message"""
    textstring = f"{name}: {vote}"
    if removalreasons:
        textstring += "\n\t"
        for reason in removalreasons:
            textstring += f" {reason},"
        textstring = textstring.rstrip(",")
    responseblock = {
        "type": "mrkdwn",
        "text": textstring
    }
    return responseblock

def build_archive_blocks(
    created_unix: str, 
    title: str,
    authorname: str,
    permalink: str,
    responseblocks: dict
) -> dict:
    """Build Slack API blocks for archive message."""
    timestamp = datetime.fromtimestamp(created_unix, tz=timezone.utc)
    timestring = f"Created {months[timestamp.month]} {ordinal(timestamp.day)} at {timestamp:%H:%M}"
    titlestring = f"<https://reddit.com{permalink}|{title}>"
    authorstring = f"Author: u/{authorname}"
    archiveblocks = [
        {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": titlestring
            }
        },
        {
            "type": "context",
            "elements": [
                {
                    "type": "plain_text",
                    "text": timestring
                }
            ]
        },
        {
            "type": "context",
            "elements": [
                {
                    "type": "mrkdwn",
                    "text": authorstring
                }
            ]
        },
        {
            "type": "divider"
        },
        {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": "*Votes on this post:*"
            }
        },
        {
            "type": "section",
            "fields": responseblocks,
        },
        {
            "type": "divider"
        },
    ]
    return archiveblocks

def rule_select_json(value: str, title:str ) -> dict:
    block_json = {
        "text": {
            "type": "plain_text",
            "text": title,
            "emoji": False,
        },
        "value": value
    }
    return block_json

def build_removal_block(rules: list) -> list:
    "Build list of Slack API multi_static_select options for removal reasons"
    removal_options = []
    for name, rule in rules.items():
        option = {
            "text": {
                "type": "plain_text",
                "text": rule["shorttext"],
                "emoji": False
            },
            "value": name
        }
        removal_options.append(option)
    return removal_options

def build_submission_block(
    created_unix: str, 
    title: str, 
    url: str, 
    authorname: str, 
    thumbnail_url: str,
    selftext: str,
    permalink: str,
    removal_options: dict
) -> dict:
    """Build Slack API blocks for new submission message."""
    # TODO Add block element for flairing posts functionality.

    # Convert PRAW object attributes to message strings
    timestamp = datetime.fromtimestamp(created_unix, tz=timezone.utc)
    timestring = f"Created {months[timestamp.month]} {ordinal(timestamp.day)}  at {timestamp:%H:%M}"
    titlestring = f"<{url}|{title}>"
    authorstring = f"Author: <https://reddit.com/u/{authorname}|u/{authorname}>"
    permalinkstring = f"https://reddit.com{permalink}"
    
    # Slack API blocks
    submissionblocks = [
        # Preamble
        {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": "<!channel> New modqueue item:"
            }
        },
        {
            "type": "divider"
        },
        # Submission info
        {
            "type": "context",
            "elements": [
                {
                    "type": "plain_text",
                    "text": timestring
                }
            ]
        },
        {
            "type": "context",
            "elements": [
                {
                    "type": "mrkdwn",
                    "text": authorstring
                }
            ]
        },
        {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": titlestring
            }
        },
        # Moderator actions
        {
            "type": "actions",
            "elements": [
                {
                    "type": "button",
                    "text": {
                        "type": "plain_text",
                        "text": "See Post",
                        "emoji": True
                    },
                    "value": "seepost",
                    "url": permalinkstring,
                    "action_id": "actionSeePost"
                }
            ]
        },
        {
            "type": "divider"
        },
        {
            "type": "actions",
            "elements": [
                {
                    "type": "radio_buttons",
                    "options": [
                        {
                            "text": {
                                "type": "plain_text",
                                "text": "Approve",
                                "emoji": True
                            },
                            "value": "+1"
                        },
                        {
                            "text": {
                                "type": "plain_text",
                                "text": "Remove",
                                "emoji": True
                            },
                            "value": "-1"
                        }
                    ],
                    "action_id": "actionApproveRemove"
                }
            ]
        },
        # Removal reasons
        {
            "type": "input",
            "element": {
                "type": "multi_static_select",
                "placeholder": {
                    "type": "plain_text",
                    "text": "Select options",
                    "emoji": True
                },
                "options": removal_options,
                "action_id": "actionRemovalReason"
            },
            "label": {
                "type": "plain_text",
                "text": "Select removal reason(s):",
                "emoji": True
            }
        },
        # Custom moderator note
		{
			"type": "input",
			"element": {
				"type": "plain_text_input",
				"multiline": True,
				"action_id": "actionModnote"
			},
			"label": {
				"type": "plain_text",
				"text": "Add modnote:",
				"emoji": True
			}
		},
        # Confirm inputs
        {
            "type": "actions",
            "elements": [
                {
                    "type": "button",
                    "text": {
                        "type": "plain_text",
                        "text": "Confirm",
                        "emoji": True
                    },
                    "value": "confirmed",
                    "action_id": "actionConfirm"
                }
            ]
        }
    ]
    
    # Handle thumbnail gracefully
    if thumbnail_url == 'self': # Replace with preview for text posts
        submissionblocks.insert(5,
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": selftext[:300] + "..." if len(selftext) > 300 else selftext
                }
            }
        )
    elif not thumbnail_url: # Leave empty if no thumbnail url given
        return submissionblocks
    else:
        submissionblocks.insert(5, 
            {
                "type": "image",
                "image_url": thumbnail_url,
                "alt_text": "thumbnail"
            }
        )
    return submissionblocks


import abc

from reddack.exceptions import (
    ActionSequenceError
)

class AbstractAction(abc.ABC):
    """Abstract class for moderation actions."""
    def __new__(cls, *args, **kwargs):
        instance = abc.ABC.__new__(cls)
        instance.timestamp = "0"
        return instance

    @abc.abstractmethod
    def update(self):
        pass

class Action(AbstractAction):
    """Empty action class for e.g buttons"""
    def update(self, action):
        pass

class State(AbstractAction):
    """Stateful action"""
    def update(self, action):
        pass

class ApproveRemove(State):
    """Mod action defining approve or remove action."""
    def __init__(self):
        self.value = None

    def update(self, state):
        self.value = state['selected_option']['value']

class RemovalReason(State):
    """Mod action defining removal reasons."""
    def __init__(self):
        self.value = []
    
    def update(self, state):
        self.value = [option['value'] for option in state['selected_options']]

class Modnote(State):
    """Mod action to add a modnote."""
    def __init__(self):
        self.value = None

    def update(self, state):
        """Retrieve modnote"""
        self.value = state['value']
        
class Confirm(Action):
    """Mod action confirming selection."""
    def __init__(self):
        self.value = False
    
    def update(self, action):
        """Confirm previous inputs"""
        self.value = True


class ReddackResponse(abc.ABC):
    """Class for storing moderator responses to Slack mod item messages"""
    
    @abc.abstractmethod
    def update(self):
        return

class SubmissionResponse(ReddackResponse):
    """Class for storing moderator responses to Slack submission messages."""
    def __init__(self, parentmsg_ts):
        self.parentmsg_ts = parentmsg_ts
        self.actions = {
            'actionConfirm' : Confirm(),
            'actionSeePost' : Action()
            }
        self.states = {
            'actionApproveRemove' : ApproveRemove(),
            'actionRemovalReason' : RemovalReason(),
            'actionModnote' : Modnote()
            }

    def update(self, request: dict, timestamp: str):
        """Update response with actions from Slack payload."""
        for action in request['actions']:
            if action['action_id'] in self.actions:
                if action['action_ts'] < self.parentmsg_ts:
                    raise ActionSequenceError(
                        "parent message", 
                        "action",
                        afterword="Something went wrong when updating responses, "
                        "if app has rebooted, try clearing known item JSON file."
                        )
                self.actions[action['action_id']].update(action)
        for blockid, blockvalue in request['state']['values'].items():
            for state in self.states:
                if state in blockvalue:
                    self.states[state].update(blockvalue[state])
                    self.states[state].timestamp = timestamp


import json
from pathlib import Path
from typing import(
    Union
)

from reddack.exceptions import ConfigError
from reddack.models import (
    Channels,
    ReddackSubmission,
    Rule,
    PrawAuth,
    SlackAuth,
    Reddack
)


def reddack_from_json(configpath: Path) -> Reddack: 
    """Load config from JSON file and return configured Reddack item"""
    with configpath.open() as f:
        masterconfig = json.load(f)
    try:
        configs = masterconfig['configs']
    except AttributeError as error:
        raise ConfigError(
            f"{error.obj!r} object is missing field {error.name!r}.",
            afterword = "Individual configurations must appear in a global 'configs' attribute."
        )
    else:
        reddacks = []
        for name, config in configs.items():
            # Create auth objects using credentials
            praw_auth = PrawAuth(
                config['praw']['auth']['client']['id'],
                config['praw']['auth']['client']['secret'],
                config['praw']['auth']['user']['name'],
                config['praw']['auth']['user']['password']
            )
            slack_auth = SlackAuth(
                config['slack']['auth']['bot_token'],
                config['slack']['auth']['user_token']
            )
            # Create channel objects for each moderation item type
            channels = {
                ReddackSubmission : Channels(
                  queue = config['slack']['channels']['submissions']['queue'],
                  archive = config['slack']['channels']['submissions']['archive']
                )
            }
            # Create rule objects and store in dictionary
            rules = {}
            for rule in config['moderation']['rules']:
                rules[rule['name']] = Rule(
                    title = rule['title'], 
                    text = rule['text'], 
                    shorttext = rule['short_text'],
                    link = rule['link'],
                    applyto = rule['applyto']
                )
            # Must have at least one type of object to moderate
            if any(c in channels for c in [ReddackSubmission]):
                pass
            else:
               raise ConfigError(
                   "Moderation requires modqueue and archive channels to be " 
                   "defined for either comments or submissions."
               )
            # Determine if non-default paths are present
            paths = {'known_items': None, 'post_dir': None}
            if ('paths' in config) and ('known_items' in config['paths']):
                paths['known_items'] = Path(config['paths']['known_items'])
            if ('paths' in config) and ('post_directory' in config['paths']):
                paths['post_dir'] = Path(config['paths']['post_directory'])
            
            reddacks.append(
                Reddack(
                    config['praw']['subreddit'],
                    praw_auth,
                    slack_auth,
                    channels,
                    rules,
                    removal_template=config['moderation']['removal_message'],
                    knownitems_path=paths['known_items'],
                    postrequest_path=paths['post_dir']
                )
            )
        return reddacks

def reddack_from_file(config_path: Union[str, Path]) -> Reddack:
    """Generate Reddack object from configuration file"""
    if config_path is str:
        config_path = Path(config_path)
    if config_path.suffix == '.json':
        return reddack_from_json(config_path)
    elif config_path.suffix == '.conf':
        # TODO Add support for .conf file types
        raise ConfigError("'.conf' file type not yet supported.")
    else:
        raise ConfigError("Configuration filetype not recognised.")


from __future__ import (
    annotations,
)

class ModFromSlackError(Exception):
    """Base class for modfromslack errors"""
    def __init__(
        self, 
        message: str,
        *,
        preamble: str | None = None,
        afterword: str | None = None
    ) -> None:
        if preamble is not None:
            message = f"{preamble} {message}"
        if afterword is not None:
            message = f"{message}\n\n{afterword}"
        super().__init__(message)

class MsgSendError(ModFromSlackError):
    """Failed to send Slack message."""

class SequenceError(ModFromSlackError):
    """Something has happened in the wrong order."""
    def __init__(
        self, 
        should_be_first, 
        should_be_second,
        *,
        preamble: str | None = None,
        afterword: str | None = None
    ) -> None:
        message = f"Expected {should_be_first} before {should_be_second}"
        super().__init__(
            message,
            preamble = preamble,
            afterword = afterword
        )

class ActionSequenceError(SequenceError):
    """App thinks action came before its parent message."""
    def __init__(
        self, 
        parentmsg_ts, 
        action_ts,
        *,
        afterword=None
    ) -> None:
        _preamble=f"'message_ts' {parentmsg_ts} is later than 'action_ts' {action_ts}"
        super().__init__(
            "parent message", 
            "action",
            preamble=_preamble, 
            afterword=afterword
        )

class ConfigError(ModFromSlackError):
    """Error in config file format."""



#!/usr/bin/env python3

"""Stub setup.py for use with legacy build tooling."""

# Future imports
from __future__ import (
    annotations,
)

# Third party imports
import setuptools

if __name__ == "__main__":
    setuptools.setup()
